<?php

/**
 * Class Expedipro_Synchro_Includes_Delivery
 *
 * This source file is subject to the MIT License
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/mit-license.php
 *
 * @category       Shipping
 * @package        Expedipro_Synchro
 * @author         Hilly
 * @copyright      Copyright (c) 2016
 * @license        http://opensource.org/licenses/mit-license.php MIT License
*/
class Expedipro_Synchro_Includes_Delivery
{
    private $statusMapping;
    private $statusState;
    private $utils;



    public function __construct( $utils = null ) {

        if ( ! $utils )
            throw new Exception('We need logger and answer management');

        $this->utils = $utils;
    }


	/**
	 * Handle the get or post request before main action
	 *
	 * @return boolean
	 */
	public function process( $action, $data )
	{
//        $this->utils->logWrite( $action );
//        $this->utils->logWrite( json_encode( $data ) );
        switch ( $action ) {
            case 'setDelivery' :
                return $this->setDelivery( $data );
            default :
                $this->utils->answerAddError('ACTPRO407', 'Unknown action within valid container', $action );
        }

        return false;
	}



    /**
	 * Extract orders and output them (to file AND to html)
	 * Note: the result page is like a log where every orderId should appears once and no more.
	 *
	 * @param string	$from mysql date	inclusive start time interval
	 * @param string	$to					inclusive end time interval
	 *
	 * @return boolean
	 */
	protected function setDelivery( $data )
    {
        /*
         * INPUT
            array(	'product' =>
                    array(	'R00' => 'Colissimo Access France',
                            'R01' => 'Colissimo Expert France',
                         ),
                    'distribution' =>
                    array(	'AVX' => 'Votre colis est remis en bureau de poste',
                            'REN' => 'Livraison effective',
                         ),
                    'carrier' =>		// not implemented
                    array(	0 => 'MyCarrier',
                            1 => 'Colissimo',
                         ),
                    'delivery' =>
                    array(	0 => array('order'=> '1', 'track'=> 'ABCDEF12', 'step' => '60', 'stepLib' => 'AVX', 'carrier' => '0', 'product' => 'R01', 'insurance' => 234, 'cost' => 15.99),
                            1 => array('order'=> '2', 'track'=> 'ZDF543A2', 'step' => '100','stepLib' => 'REN', 'carrier' => '1', 'product' => 'R00','insurance' => 0, 'cost' => 8.22),
                         )
                )

            OUTPUT
             array( type = delivery,
                    id = 9999
                    array(  'tracking'  => unknown | new | known | grouped
                            'status'    => first | stand
                    )
             )
        */
        if (! isset ( $data['delivery']) ) {
            $this->utils->answerAddError('ACTDEL001', 'delivery', 'no delivery data');
            return false;
        }

        // dataDelivery has a 0 based index while orderArray is orderId indexed
        $ordersArray = array_flip( $this->utils->array_column_54( $data['delivery'], 'order') );
//        $max = count($ordersArray);

        try {
            // https://github.com/OpenMage/magento-mirror/blob/magento-1.7/lib/Varien/Data/Collection/Db.php
            $collection = Mage::getModel('sales/order')
                ->getCollection()
                ->addAttributeToFilter('increment_id', array('in' => array_keys($ordersArray)) );
//var_dump((string)$collection->getSelect());
            if ( ! $collection ) {
                $this->utils->answerAddError('ACTDEL402', 'No order match in DB', print_r(array_keys( $ordersArray )) );
                return false;
            }
        } catch (Exception $e) {
            $this->utils->answerAddError('ACTDEL402', 'Can not access DB tables', $e->getMessage() );
            return false;
        }

        // initiate mapping between steps and status
        $this->setStatusMapping( $this->utils->getConfiguration('deliveryStart'), $this->utils->getConfiguration('deliveryDone') );

        foreach( $collection as $order )
        {
            try {
                $currentDelivery = $data['delivery'][ $ordersArray[ $order->getData('increment_id') ] ];

                // unhold to change status
                if ( $order->canUnhold() ) {
                    $order->unhold()
                          ->save();
                }

                // create shipment / add tracking / add comment
                if ( ! $this->manageShipment( $order, $currentDelivery, $data['product'], $data['distribution'] ) )
                    continue;


                // Important, we rely on status not on state
                $status = $this->getStatusMapping( $currentDelivery['step'] );

                // Avoid creating order history events when status does not change
                if ( $status == $order->getStatus() ) {
                    $this->utils->answerAddData( $order->getData('increment_id'), 'delivery', array('status' => 'stand') );
                    continue;
                }

                $this->updateStatus( $order, $status, $currentDelivery, $data['product'] );
                $this->utils->answerAddData( $order->getData('increment_id'), 'delivery', array('status' => 'update') );
            }
            catch (Exception $e)
            {
                $this->utils->answerAddError('ACTDEL403', 'Can not update order incrementId '. $order->getData('increment_id'), $e->getMessage() );
            }
        }

//      $this->utils->logWrite( $this->utils->getAnswer() );
        return count( $collection );
    }


    /**
     * Handle shipment object for the order
     *
     * @param $order
     * @param $delivery
     * @param $product
     * @param $distribution
     *
     * @return false | Mage_Sales_Model_Order_Shipment
     */
    private function manageShipment( $order, $delivery, $product, $distribution ) {

        // get or assign a new shipment to the order
        if ( ! $shipment = $this->getShipment( $order, $delivery['track'] ) ) {
            $this->utils->answerAddData( $order->getData('increment_id'), 'delivery', array('shipment' => 'error') );
            return false;
        }

        // add one or many comment / track number
        return $this->updateShipment( $shipment, $delivery, $product, $distribution );
    }


    /**
     * get an existing or new shipment
     *
     * @param Mage_Sales_Model_Order $order
     *
     * @return false | @param Mage_Sales_Model_Order_Shipment
     */
    private function getShipment( $order, $trackNumber ='' )
    {
/*      // can not rely on canShip() that fails when no more items to ship
        // since we can update several time an shipment and order state
        if ( ! $order->canShip() )
            return false;
*/
        //Check to see if the order already shipments
        if( $order->hasShipments() ) {
//            $this->utils->answerAddData( $order->getData('increment_id'), 'delivery', array('shipment' => 'exist') );
            $shipments = $order->getShipmentsCollection();
            foreach ( $shipments as $one ) {
                if ( $this->isTrackExist( $one, $trackNumber ) )
                    return $one;
            }
            return $shipments->getFirstItem();
        }

        // http://magento.stackexchange.com/questions/3252/programmatically-creating-shipments
        // get a Mage_Sales_Model_Order_Shipment object
        $convert  = Mage::getModel('sales/convert_order');
        $shipment = $convert->toShipment($order);

        // add items to shipment
        foreach ($order->getAllVisibleItems() as $orderItem)
        {
            if ( ! $orderItem->getQtyToShip()
                OR $orderItem->getIsVirtual())
                continue;

            $item = $convert->itemToShipmentItem($orderItem);
            $qty = $orderItem->getQtyToShip();
            $item->setQty($qty);
            $shipment->addItem($item);
        }
        $shipment->register();

        $transaction = Mage::getModel('core/resource_transaction')
            ->addObject($shipment)
            ->addObject($shipment->getOrder())
            ->save();

        // Set order to processing (Magento constraint order with shipment must be at least processing)
        $order->setIsInProcess(true) // TODO : parameter = notify customer ? True required to moved to processing
              ->save();

//        $this->utils->answerAddData( $order->getData('increment_id'), 'delivery', array('shipment' => 'new') );
        return $shipment;
    }


    /**
     * Update shipment object with tracking number and comments
     *
     * @param Mage_Sales_Model_Order_Shipment $shipment
     * @param array     $delivery       delivery from Expedipro
     * @param array     $products       products from Expedipro
     * @param array     $distributionLabel   label from Expedipro
     *
     * @return Mage_Sales_Model_Order_Shipment
     */
    private function updateShipment( $shipment, $delivery= array(), $products = array(), $distributionLabel = array())
    {
        // add tracking info if was not yet defined
        if ( ! $trackExist = $this->isTrackExist($shipment, $delivery['track']))
        {
            $track = Mage::getModel('sales/order_shipment_track')
                ->setNumber($delivery['track'])
                ->setCarrierCode('expedipro')
                ->setTitle($products[ $delivery['product'] ]);
            $shipment->addTrack($track);

            // Add full shipment description
            $comment = $products[ $delivery['product'] ] .
                ($delivery['insurance'] ? ', assuré pour ' . str_replace('.', ',', $delivery['insurance']) . '€' : '') .
                ($delivery['cost'] ? ', coût ht = ' . str_replace('.', ',', $delivery['cost']) : '') .
                ' Suivi sur ' . $this->utils->getConfiguration('deliveryTracking', 'expedipro') . $delivery['track'];
            $shipment->addComment($comment, false);
        }
        else
        {
            // Add distribution progress
            $comment = $delivery['track'] . ' : ' . $distributionLabel[ $delivery['stepLib'] ];
            $shipment->addComment($comment, false);
        }
        // note updated $shipment object is availbale in caller
        return $shipment->save();
    }


    /**
     * Search a tracking number into a shipment object
     *
     * @param Mage_Sales_Model_Order_Shipment   $shipment
     * @param string                            $trackNumber
     *
     * @return bool
     */
    private function isTrackExist( $shipment, $trackNumber ='') {
        if ( ! ( $trackNumber AND $allTracks = $shipment->getAllTracks() ) )
            return false;

        foreach ( $allTracks as $oneTrack ) {
            if ( $oneTrack->getTrackNumber() == $trackNumber )
                return true;
        }
        return false;
    }





    /**
     * Manage status history and comments
     *
     * @param Mage_Sales_Model_Order $order
     * @param string    magento status to change to (nb. can be null)
     * @param array     $delivery   details from Expedipro
     * @param array     $products   details from Expedipro
     *
     * @return bool|int
     */
    private function updateStatus( $order, $status = null, $delivery= array(), $products = array())
    {
//        $status = 'processing';
        // @todo Mage_Sales_Model_Order_Shipment_Api already does what we do here !
        // Add a private message for this Order to display parcel details on the order back office page
        $comment = $products[ $delivery['product'] ].
            ( $delivery['insurance'] ? ', assuré pour ' . str_replace('.', ',',$delivery['insurance']) . '€' : '') .
            ( $delivery['cost'] ? ', coût ht = ' . str_replace('.', ',',$delivery['cost']) : '') .
            'Suivi sur ' . $this->utils->getConfiguration('deliveryTracking', 'expedipro') . $delivery['track'];

        if ( null === $status ) {
            return $order->addStatusHistoryComment( $comment, false)  // does not change status
                         ->save();
        }

        // https://magento2.atlassian.net/wiki/display/m1wiki/Order+management+schema?preview=/14024877/14024981/order_management_in_magento.png
        // STATES rules
        //  - order remains "processing" until all item qty has been shipped
        //  - order changes to "complete" once all item qty had been shipped and an invoice is generated
        //  - order changes to "closed" once all credit had been paid

        $state = $this->getStatusState( $status );
        // STATE_COMPLETE || STATE_CLOSED
        if ( $order->isStateProtected( $state ) )
            $this->fulfillProtectedStates( $order ); // generate an historyComment

        // duplicate from protected function $order->_setState()
        // required to bypass $order->setState( $shouldProtectState )
        return $order->setData('state', $state)
                     ->addStatusHistoryComment( $comment, $status)  // change status with a comment
                     ->setStatus( $status )
                     ->save();
    }


    /**
    * Retrieve state available for status
    * Get first assigned state for specified status
    * note : check Mage_Sales_Model_Order_Config
    *
    * @param string $status
     *
    * @return string
    */
    private function getStatusState($status = '')
    {
        if ( isset( $this->statusState[ $status ]) )
            return $this->statusState[ $status ];

        $collection = Mage::getResourceModel('sales/order_status_collection')->addStatusFilter($status);
        foreach ($collection as $state) {
//            var_dump ( $state );
            $this->statusState[ $status ] = $state->getData('state');
            return $this->statusState[ $status ];
        }
        return '';
    }


    /**
     * Magento protects state complete | close
     * Some actions are required before changing state to those
     *
     * @param $order
     */
    private function fulfillProtectedStates( $order ) {
        // invoice the order
        if ($order->hasInvoices())
        {
            // select the last invoice to attach the note to
            $invoice = $order->getInvoiceCollection()->getLastItem();
        }
        else
        {
            // prepare a brand-new invoice
            $invoice = $order->prepareInvoice();
            $invoice->register();
        }

        // capture the invoice if possible
        if ($invoice->canCapture())
            $invoice->Capture();

        // some magento versions prevent multiple pay() calls from have impact,
        // but others don't.  If pay is called multiple times, Order.Total Paid is off.
        if ($invoice->getState() != Mage_Sales_Model_Order_Invoice::STATE_PAID)
            $invoice->pay();

        // save the new invoice
        $transactionSave = Mage::getModel('core/resource_transaction')
            ->addObject($invoice)
            ->addObject($invoice->getOrder());
        $transactionSave->save();
        return true;
    }


    /**
     * Get closest magento status for an Expedipro step
     *
     * @param int $step     delivery step from expedipro
     *
     * @return array|null
     */
    private function getStatusMapping( $step = 0)
	{
		if ( ! $step = (int) $step )
			return null;

        $previous = null;
        foreach ($this->statusMapping AS $expedipro => $magento )
        {
            $expedipro = (int) $expedipro;
            if ( $step === $expedipro )
                return $magento;
            if ( $step  <  $expedipro )
                return $previous;	// if expedipro send a unknown code, don't break and send the closest one

            $previous = $magento;
        }
        return null; // $previous
    }

    /**
     * Set mapping between expedipro steps and magento status (cf. system.xml)
     *
     * @param string $start
     * @param string $done
     */
    private function setStatusMapping($start = '', $done = '' )
    {
        $this->statusMapping =
            array (
            1 => null,  // to early in the process to trigger email to customer
            10=> $start,
            11=> $start,
            20=> $start,
            30=> $start,
            40=> $start,
            50=> $start,
            60=> $start,
            70=> $start,
            80=> $start,
            90=> $start,
            100=>$done
        );
    }

}
?>